import os
import json
import time
import redash

from redash_stmo import settings
from redash_stmo.resources import add_resource

from flask import jsonify
from flask_login import login_required

from celery.utils.log import get_task_logger
from redash import models, redis_connection, statsd_client
from redash.worker import celery
from redash.utils import parse_human_time
from redash.monitor import get_status as original_get_status
from redash.query_runner import BaseQueryRunner
from redash.handlers.base import routes, BaseResource
from redash.permissions import require_super_admin


logger = get_task_logger(__name__)


class DataSourceHealthResource(BaseResource):
    def get(self):
        health_data = json.loads(redis_connection.get('data_sources:health') or '{}')
        return jsonify(health_data)

def store_health_status(data_source_id, data_source_name, query_text, data):
    key = "data_sources:health"

    cache = json.loads(redis_connection.get(key) or '{}')
    if data_source_id not in cache:
        cache[data_source_id] = {
            "metadata": {"name": data_source_name},
            "queries": {}
        }
    cache[data_source_id]["queries"][query_text] = data

    cache[data_source_id]["status"] = "SUCCESS"
    for query_status in cache[data_source_id]["queries"].values():
        if query_status["status"] == "FAIL":
            cache[data_source_id]["status"] = "FAIL"
            break

    redis_connection.set(key, json.dumps(cache))


@celery.task(name="redash_stmo.health.update_health_status", time_limit=90, soft_time_limit=60)
def update_health_status():
    for data_source in models.DataSource.query:
        logger.info(u"task=update_health_status state=start ds_id=%s", data_source.id)

        runtime = None
        query_text = data_source.query_runner.noop_query
        ds_id = str(data_source.id)

        custom_query_env_var = "REDASH_CUSTOM_HEALTH_QUERIES_{data_source_id}".format(data_source_id=ds_id)
        custom_query = os.environ.get(custom_query_env_var, "")
        query_text = custom_query or query_text

        try:
            start_time = time.time()
            test_connection(data_source.query_runner, query_text)
            runtime = time.time() - start_time
        except NotImplementedError:
            logger.info(u"Unable to compute health status without test query for %s", data_source.name)
            continue
        except Exception as e:
            logger.warning(u"Failed health check for the data source: %s", data_source.name, exc_info=1)
            statsd_client.incr('update_health_status.error')
            logger.info(u"task=update_health_status state=error ds_id=%s runtime=%.2f", data_source.id, time.time() - start_time)

        status = {
            "status": "FAIL" if runtime is None else "SUCCESS",
            "last_run": start_time,
            "last_run_human": str(parse_human_time(str(start_time))),
            "runtime": runtime
        }
        store_health_status(ds_id, data_source.name, query_text, status)


def test_connection(self, custom_query_text=None):
    if self.noop_query is None:
        raise NotImplementedError()

    query_text = custom_query_text or self.noop_query
    data, error = self.run_query(query_text, None)

    if error is not None:
        raise Exception(error)

@login_required
@require_super_admin
def stmo_status_api():
    status = original_get_status()
    health_data = json.loads(redis_connection.get('data_sources:health') or '{}')

    # Get the top level status for each data source
    for health_data_point in health_data.values():
        data_source_name = health_data_point["metadata"]["name"]
        dashboard_label = "[Data Source Health] {name}".format(name=data_source_name)
        status[dashboard_label] = health_data_point["status"]
    return jsonify(status)


def health(app=None):
    """A Redash extension to add datasource health status reporting."""

    # Override the default status API view with our extended view
    app.view_functions['%s.status_api' % routes.name] = stmo_status_api

    # Add a new endpoint with full health data
    add_resource(app, DataSourceHealthResource, '/status/data_sources/health.json')

    # Add the update_health_status task to a list of periodic tasks
    if not hasattr(app, 'periodic_tasks'):
        app.periodic_tasks = {}

    app.periodic_tasks['update_health_status'] = {
        'sig': update_health_status.s(),
        'schedule': settings.HEALTH_QUERIES_REFRESH_SCHEDULE,
    }



import json
import mock

from tests import BaseTestCase
from flask import Flask

from redash.models import DataSource
from redash.query_runner.pg import PostgreSQL
from redash_stmo.data_sources.version import version


class TestDatasourceVersion(BaseTestCase):
    EXPECTED_DOC_URL = "www.example.com"
    def setUp(self):
        super(TestDatasourceVersion, self).setUp()
        self.admin = self.factory.create_admin()
        self.data_source = self.factory.create_data_source()
        self.patched_run_query = self._setup_mock('redash.query_runner.pg.PostgreSQL.run_query')
        self.patched_runner_type = self._setup_mock('redash.query_runner.pg.PostgreSQL.type')
        version(self.app)

    def _setup_mock(self, function_to_patch):
        patcher = mock.patch(function_to_patch)
        patched_function = patcher.start()
        self.addCleanup(patcher.stop)
        return patched_function

    def _test_expected_version_returned(self, expected_version, version_string, runner_type):
        self.patched_runner_type.return_value = runner_type
        self.patched_run_query.return_value = (json.dumps({
            "rows":
                [{ "version": version_string.format(version=expected_version) }]
        }), None)
        rv = self.make_request('get', '/api/data_sources/{}/version'.format(self.data_source.id), user=self.admin)
        self.assertEqual(200, rv.status_code)
        self.assertEqual(rv.json['version'], expected_version)

    def test_gets_postgres_version(self):
        RUNNER_TYPE = "pg"
        DATASOURCE_VERSION = "9.5.10"
        VERSION_STRING = (
            "PostgreSQL {version} on x86_64-pc-linux-gnu, compiled by gcc "
            "(GCC) 4.8.3 20140911 (Red Hat 4.8.3-9), 64-bit"
        )
        self._test_expected_version_returned(DATASOURCE_VERSION, VERSION_STRING, RUNNER_TYPE)

    def test_gets_redshift_version(self):
        RUNNER_TYPE = "redshift"
        DATASOURCE_VERSION = "1.0.3688"
        VERSION_STRING = (
            "PostgreSQL 8.0.2 on i686-pc-linux-gnu, compiled by GCC "
            "gcc (GCC) 3.4.2 20041017 (Red Hat 3.4.2-6.fc3), Redshift {version}"
        )
        self._test_expected_version_returned(DATASOURCE_VERSION, VERSION_STRING, RUNNER_TYPE)

    def test_gets_mysql_version(self):
        RUNNER_TYPE = "mysql"
        DATASOURCE_VERSION = "5.7.16"
        VERSION_STRING = "{version}-log"
        self._test_expected_version_returned(DATASOURCE_VERSION, VERSION_STRING, RUNNER_TYPE)

    def test_unexpected_json(self):
        self.patched_runner_type.return_value = "pg"
        self.patched_run_query.return_value = (json.dumps({
            "rows":
                [{ "bad_json": "foo" }]
        }), None)
        rv = self.make_request('get', '/api/data_sources/{}/version'.format(self.data_source.id), user=self.admin)
        self.assertEqual(200, rv.status_code)
        self.assertEqual(rv.json['version'], None)



import mock

from tests import BaseTestCase
from flask import Flask

from redash.models import DataSource
from redash.query_runner.pg import PostgreSQL
from redash_stmo.data_sources.link import link


class TestDatasourceLink(BaseTestCase):
    EXPECTED_DOC_URL = "www.example.com"
    def setUp(self):
        super(TestDatasourceLink, self).setUp()
        self.patched_query_runners = self._setup_mock('redash_stmo.data_sources.link.query_runners')
        self.patched_query_runners.return_value = {}
        link(self.app)

    def _setup_mock(self, function_to_patch):
        patcher = mock.patch(function_to_patch)
        patched_function = patcher.start()
        self.addCleanup(patcher.stop)
        return patched_function

    def test_gets_datasource_link_and_type(self):
        admin = self.factory.create_admin()
        data_source = self.factory.create_data_source()
        data_source.options["doc_url"] = self.EXPECTED_DOC_URL

        rv = self.make_request('get', '/api/data_sources/{}/link'.format(data_source.id), user=admin)
        self.assertEqual(200, rv.status_code)
        self.assertEqual(rv.json['message']['type_name'], data_source.query_runner.name())
        self.assertEqual(rv.json['message']["doc_url"], self.EXPECTED_DOC_URL)



from redash.handlers import api

def add_resource(app, resource, endpoint):
    """
    After api.init_app() is called, api.app should be set by Flask (but it's not) so that
    further calls to add_resource() are handled immediately for the given app.
    """
    api.app = app
    api.add_org_resource(resource, endpoint)



import os
import json
import mock

from tests import BaseTestCase
from flask import Flask

from redash import redis_connection
from redash_stmo.data_sources.health import store_health_status, update_health_status, health



class TestHealthStatus(BaseTestCase):
    def setUp(self):
        super(TestHealthStatus, self).setUp()
        self.patched_store_health_status = self._setup_mock('redash_stmo.data_sources.health.store_health_status')
        self.patched_run_query = self._setup_mock('redash.query_runner.pg.PostgreSQL.run_query')

        self.patched_run_query.return_value = ("some_data", None)
        os.environ["REDASH_CUSTOM_HEALTH_QUERIES_1"] = ""

        health(self.app)

    def _setup_mock(self, function_to_patch):
        patcher = mock.patch(function_to_patch)
        patched_function = patcher.start()
        self.addCleanup(patcher.stop)
        return patched_function

    def test_store_health_status_sets_correct_keys(self):
        current_health = redis_connection.get('data_sources:health')
        self.assertEqual(None, current_health)

        DATA_SOURCE = self.factory.create_data_source()
        QUERY_SUCCESS = "SELECT 1"
        QUERY_FAIL = "SELECT meep"
        SOME_DATA_FAIL = {"a": "b", "foo": "bar", "status": "FAIL"}
        SOME_DATA_SUCCESS = {"a": "b", "foo": "bar", "status": "SUCCESS"}
        store_health_status(str(DATA_SOURCE.id), DATA_SOURCE.name, QUERY_FAIL, SOME_DATA_FAIL)
        store_health_status(str(DATA_SOURCE.id), DATA_SOURCE.name, QUERY_SUCCESS, SOME_DATA_SUCCESS)

        '''
          The expected format of the cached health data is:
          {
            "<data_source_id>": {
              "metadata": "<data_source_name>",
              "queries": {
                "<query_text>": {...},
                "<query_text>": {...},
                "<query_text>": {...},
                ...
              }
            },
            ...
          }
        '''

        current_health = json.loads(redis_connection.get('data_sources:health'))

        # There is 1 data source.
        self.assertEqual(1, len(current_health.keys()))
        self.assertEqual(DATA_SOURCE.id, int(current_health.keys()[0]))

        # The data source has "metadata", "queries" and "status" keys.
        ds_id = str(DATA_SOURCE.id)
        self.assertEqual(3, len(current_health[ds_id].keys()))
        self.assertTrue("metadata" in current_health[ds_id].keys())
        self.assertTrue("queries" in current_health[ds_id].keys())
        self.assertTrue("status" in current_health[ds_id].keys())

        # There are two queries with correct data
        self.assertEqual(2, len(current_health[ds_id]["queries"]))
        self.assertTrue(QUERY_SUCCESS in current_health[ds_id]["queries"].keys())
        self.assertTrue(QUERY_FAIL in current_health[ds_id]["queries"].keys())
        self.assertEqual(SOME_DATA_FAIL, current_health[ds_id]["queries"][QUERY_FAIL])
        self.assertEqual(SOME_DATA_SUCCESS, current_health[ds_id]["queries"][QUERY_SUCCESS])
        self.assertEqual(SOME_DATA_FAIL["status"], current_health[ds_id]["status"])

    def test_health_status_success(self):
        data_sources = []
        for i in range(5):
            data_sources.append(self.factory.create_data_source())

        update_health_status()

        # Status is updated for each of the 5 data sources
        self.assertEqual(self.patched_store_health_status.call_count, 5)

        # The data source name and id is correctly passed in the last call of update_health_status()
        args, kwargs = self.patched_store_health_status.call_args
        self.assertEqual(str(data_sources[-1].id), args[0])
        self.assertEqual(data_sources[-1].name, args[1])

        # All expected status keys are available.
        EXPECTED_KEYS = ["status", "last_run", "last_run_human", "runtime"]
        NEW_STATUS = args[3]
        new_status_keys = set(NEW_STATUS.keys())
        self.assertEqual(set(EXPECTED_KEYS), new_status_keys)

        self.assertEqual("SUCCESS", NEW_STATUS["status"])
        for key in EXPECTED_KEYS[1:]:
            self.assertIsNotNone(NEW_STATUS[key])

    def test_health_status_run_query_throws_noop_not_implemented_exception(self):
        data_source = self.factory.create_data_source()

        def exception_raiser(*args, **kwargs):
          raise NotImplementedError

        self.patched_run_query.side_effect = exception_raiser
        update_health_status()

        # Status is updated for the one data source
        self.assertEqual(self.patched_store_health_status.call_count, 0)

    def test_health_status_run_query_throws_exception(self):
        data_source = self.factory.create_data_source()

        def exception_raiser(*args, **kwargs):
          raise Exception

        self.patched_run_query.side_effect = exception_raiser
        update_health_status()

        # Status is updated for the one data source
        self.assertEqual(self.patched_store_health_status.call_count, 1)

        # The data source name is correctly passed in the last call of update_health_status()
        args, kwargs = self.patched_store_health_status.call_args
        self.assertEqual(str(data_source.id), args[0])
        self.assertEqual(data_source.name, args[1])
        self.assertEqual(data_source.query_runner.noop_query, args[2])

        # All expected status keys are available.
        EXPECTED_KEYS = ['status', 'last_run', 'last_run_human', 'runtime']
        NEW_STATUS = args[3]
        new_status_keys = set(NEW_STATUS.keys())
        self.assertEqual(set(EXPECTED_KEYS), new_status_keys)

        self.assertEqual('FAIL', NEW_STATUS['status'])
        self.assertIsNotNone(NEW_STATUS['last_run'])
        self.assertIsNotNone(NEW_STATUS['last_run_human'])
        self.assertIsNone(NEW_STATUS['runtime'])

    def test_health_status_custom_query(self):
        CUSTOM_QUERY = "select * from table"
        data_source = self.factory.create_data_source()
        os.environ["REDASH_CUSTOM_HEALTH_QUERIES_1"] = CUSTOM_QUERY

        update_health_status()

        args, kwargs = self.patched_store_health_status.call_args
        self.assertNotEqual(data_source.query_runner.noop_query, args[2])
        self.assertEqual(CUSTOM_QUERY, args[2])



"""
A custom Redash query runner for Mozilla's ActiveData service:

More information:

- https://wiki.mozilla.org/EngineeringProductivity/Projects/ActiveData
- https://github.com/mozilla/ActiveData/blob/dev/docs/redash.md

Originally written by Github user @klahnakoski
Original link: https://github.com/klahnakoski/ActiveData-redash-query-runner/blob/c0e7286c09c6f1eb6746a6c7cca581bea79f4757/active_data.py
"""
import json
import logging

import requests

from redash.query_runner import (TYPE_INTEGER, TYPE_STRING, TYPE_FLOAT,
                                 BaseSQLQueryRunner, register)
from redash.utils import JSONEncoder


logger = logging.getLogger(__name__)

TYPES_MAP = {
    bool: TYPE_INTEGER,
    str: TYPE_STRING,
    unicode: TYPE_STRING,
    dict: TYPE_STRING,
    list: TYPE_STRING,
    int: TYPE_INTEGER,
    long: TYPE_INTEGER,
    float: TYPE_FLOAT,
    "string": TYPE_STRING,
    "object": TYPE_STRING,
    "long": TYPE_STRING,
    "double": TYPE_FLOAT,
    "integer": TYPE_FLOAT
}


class ActiveData(BaseSQLQueryRunner):
    noop_query = "SELECT 1"

    @classmethod
    def configuration_schema(cls):
        return {
            "type": "object",
            "properties": {
                "host_url": {
                    "type": "string",
                    "title": "Host URL",
                    "default": "https://activedata.allizom.org:80",
                    "info": "Please include a port. Do not end with a trailing slash."
                },
                "doc_url": {
                    "type": "string",
                    "title": "Documentation URL",
                    "default": "https://github.com/klahnakoski/ActiveData/tree/dev/docs"
                },
                "toggle_table_string": {
                    "type": "string",
                    "title": "Toggle Table String",
                    "default": "_v",
                    "info": "This string will be used to toggle visibility of tables in the schema browser when editing a query in order to remove non-useful tables from sight."
                }
            },
            "required": ["host_url"]
        }

    @classmethod
    def name(cls):
        return "ActiveData"

    @classmethod
    def type(cls):
        return "activedata"

    @classmethod
    def enabled(cls):
        return True

    def _get_tables(self, schema):
        query = {
            "from": "meta.columns",
            "select": [
                "name",
                "type",
                "table"
            ],
            "where": {"not": {"prefix": {"es_index": "meta."}}},
            "limit": 1000,
            "format": "list"
        }
        results = self.run_jx_query(query, None)

        for row in results['data']:
            table_name = row['table']

            if table_name not in schema:
                schema[table_name] = {'name': table_name, 'columns': []}

            schema[table_name]['columns'].append(
                row['name'] + ' (' + TYPES_MAP.get(row['type'], TYPE_STRING) + ')'
            )

        return [
            {
                'name': table['name'],
                'columns': sorted(table['columns'])
            }
            for table in schema.values()
        ]

    def run_jx_query(self, query, user):
        data = json.dumps(query, ensure_ascii=False)
        result = requests.post(
            self.configuration['host_url'] + "/query",
            data=data,
        )
        response = json.loads(result.content)

        if response.get('type') == "ERROR":
            cause = self.find_error_cause(response)
            raise Exception(cause)
        return response

    def run_query(self, annotated_query, user):
        request = {}
        comment, request["sql"] = annotated_query.split("*/", 2)
        meta = request['meta'] = {}
        for kv in comment.strip()[2:].split(","):
            k, v = [s.strip() for s in kv.split(':')]
            meta[k] = v

        logger.debug("Send ActiveData a SQL query: %s", request['sql'])
        data = json.dumps(request, ensure_ascii=False)
        result = requests.post(
            self.configuration['host_url'] + "/sql",
            data=data,
        )
        response = json.loads(result.content)

        if response.get('type') == "ERROR":
            cause = self.find_error_cause(response)
            return None, cause

        output = self.normalize_response(response)
        json_data = json.dumps(output, cls=JSONEncoder)
        return json_data, None

    def normalize_response(self, table):
        columns = {}  # MAP FROM name TO (MAP FROM type TO (full_name))
        output = []

        def get_unique_name(name, type):
            all_types = columns.get(name)
            if all_types is None:
                all_types = columns[name] = {}
            specific_type = all_types.get(type)
            if specific_type is None:
                if all_types:
                    specific_type = all_types[type] = name + "." + type
                else:
                    specific_type = all_types[type] = name
            return specific_type

        for r in table['data']:
            new_row = {}
            for i, cname in enumerate(table['header']):
                val = r[i]
                if val is None:
                    continue
                if isinstance(val, (dict, list)):
                    val = json.dumps(val, cls=JSONEncoder)
                col = get_unique_name(cname, TYPES_MAP.get(type(val), TYPE_STRING))
                new_row[col] = val
            output.append(new_row)

        output_columns = [
            {
                "name": full_name,
                "type": ctype,
                "friendly_name": full_name
            }
            for cname, types in columns.items()
            for ctype, full_name in types.items()
        ]

        return {
            'columns': output_columns,
            'rows': output
        }

    def find_error_cause(self, response):
        while response.get('cause') is not None:
            cause = response['cause']
            if isinstance(cause, list):
                response = cause[0]
            else:
                response = cause
        return response.get('template')


register(ActiveData)



from redash_stmo.resources import add_resource

from redash.models import DataSource
from redash.handlers.base import BaseResource, get_object_or_404
from redash.permissions import require_access, view_only
from redash.query_runner import query_runners

DATASOURCE_URLS = {
    "bigquery": "https://cloud.google.com/bigquery/docs/reference/legacy-sql",
    "Cassandra": "http://cassandra.apache.org/doc/latest/cql/index.html",
    "dynamodb_sql": "https://dql.readthedocs.io/en/latest/",
    "baseelasticsearch": "https://www.elastic.co/guide/en/elasticsearch/reference/current/index.html",
    "google_spreadsheets": "http://redash.readthedocs.io/en/latest/datasources.html#google-spreadsheets",
    "hive": "https://cwiki.apache.org/confluence/display/Hive/LanguageManual",
    "impala": "http://www.cloudera.com/documentation/enterprise/latest/topics/impala_langref.html",
    "influxdb": "https://docs.influxdata.com/influxdb/v1.0/query_language/spec/",
    "jirajql": "https://confluence.atlassian.com/jirasoftwarecloud/advanced-searching-764478330.html",
    "mongodb": "https://docs.mongodb.com/manual/reference/operator/query/",
    "mssql": "https://msdn.microsoft.com/en-us/library/bb510741.aspx",
    "mysql": "https://dev.mysql.com/doc/refman/5.7/en/",
    "oracle": "http://docs.oracle.com/database/121/SQLRF/toc.htm",
    "pg": "https://www.postgresql.org/docs/current/",
    "redshift": "http://docs.aws.amazon.com/redshift/latest/dg/cm_chap_SQLCommandRef.html",
    "presto": "https://prestodb.io/docs/current/",
    "python": "http://redash.readthedocs.io/en/latest/datasources.html#python",
    "insecure_script": "http://redash.readthedocs.io/en/latest/datasources.html#python",
    "sqlite": "http://sqlite.org/lang.html",
    "treasuredata": "https://docs.treasuredata.com/categories/hive",
    "url": "http://redash.readthedocs.io/en/latest/datasources.html#url",
    "vertica": (
        "https://my.vertica.com/docs/8.0.x/HTML/index.htm#Authoring/"
        "ConceptsGuide/Other/SQLOverview.htm%3FTocPath%3DSQL"
        "%2520Reference%2520Manual%7C_____1"
    )
}


class DataSourceLinkResource(BaseResource):
    def get(self, data_source_id):
        data_source = get_object_or_404(
            DataSource.get_by_id_and_org,
            data_source_id,
            self.current_org,
        )
        require_access(data_source.groups, self.current_user, view_only)
        try:
            result = {
                "type_name": data_source.query_runner.name(),
                "doc_url": data_source.options.get("doc_url", None)
            }
        except Exception as e:
            return {"message": unicode(e), "ok": False}
        else:
            return {"message": result, "ok": True}

def link(app=None):
    for runner_type, runner_class in query_runners.items():
        if runner_type not in DATASOURCE_URLS:
            continue

        runner_class.add_configuration_property("doc_url", {
            "type": "string",
            "title": "Documentation URL",
            "default": DATASOURCE_URLS[runner_type]})

    add_resource(app, DataSourceLinkResource, '/api/data_sources/<data_source_id>/link')



import json
import logging

from redash_stmo.resources import add_resource

from redash.models import DataSource
from redash.handlers.base import BaseResource, get_object_or_404
from redash.permissions import require_access, view_only

logger = logging.getLogger(__name__)


DATASOURCE_VERSION_PARSE_INFO = {
    "pg": {
        "version_query": "select version()",
        "delimiter": " ",
        "index": 1
    },
    "redshift": {
        "version_query": "select version()",
        "delimiter": " ",
        "index": -1
    },
    "mysql": {
        "version_query": "select version()",
        "delimiter": "-",
        "index": 0
    }
}

class DataSourceVersionResource(BaseResource):
    def get(self, data_source_id):
        data_source = get_object_or_404(
            DataSource.get_by_id_and_org,
            data_source_id,
            self.current_org
        )
        require_access(data_source.groups, self.current_user, view_only)
        version_info = get_data_source_version(data_source.query_runner)
        return {"version": version_info}

def get_data_source_version(query_runner):
    parse_info = DATASOURCE_VERSION_PARSE_INFO.get(query_runner.type())
    if parse_info is None:
        return None

    data, error = query_runner.run_query(parse_info["version_query"], None)
    if error is not None:
        logger.error(
            "Unable to run version query for %s: %s", query_runner.type(), error)
        return None
    try:
        version = json.loads(data)['rows'][0]['version']
    except (KeyError, IndexError) as err:
        logger.exception(
            "Unable to parse data source version for %s: %s", query_runner.type(), err)
        return None

    version = version.split(parse_info["delimiter"])[parse_info["index"]]
    return version

def version(app=None):
    add_resource(app, DataSourceVersionResource, '/api/data_sources/<data_source_id>/version')
